from contextvars import ContextVar
from typing import Any

from pyxxl.schema import RunData


_global_vars: ContextVar[dict] = ContextVar("pyxxl_vars", default={})


class GlobalVars:
    @staticmethod
    def _set_var(name: str, obj: Any) -> None:
        _global_vars.get()[name] = obj

    @staticmethod
    def _get_var(name: str) -> Any:
        return _global_vars.get()[name]

    @staticmethod
    def set_xxl_run_data(data: RunData) -> None:
        GlobalVars._set_var("xxl_kwargs", data)

    @property
    def xxl_run_data(self) -> RunData:
        return self._get_var("xxl_kwargs")


g = GlobalVars()



import logging

from aiohttp import web

from pyxxl import error
from pyxxl.schema import RunData


logger = logging.getLogger("pyxxl")

routes = web.RouteTableDef()


@routes.post("/beat")
async def beat(request: web.Request) -> web.Response:
    logger.debug("beat")
    return web.json_response(dict(code=200, msg=None))


@routes.post("/idleBeat")
async def idle_beat(request: web.Request) -> web.Response:
    data = await request.json()
    job_id = data["jobId"]
    logger.debug("idleBeat: %s" % data)
    if await request.app["executor"].is_running(data["jobId"]):
        return web.json_response(dict(code=500, msg="job %s is running." % job_id))
    return web.json_response(dict(code=200, msg=None))


@routes.post("/run")
async def run(request: web.Request) -> web.Response:
    """
        {
        "jobId":1,                                  // 任务ID
        "executorHandler":"demoJobHandler",         // 任务标识
        "executorParams":"demoJobHandler",          // 任务参数
        "executorBlockStrategy":"COVER_EARLY",      // 任务阻塞策略，可选值参考 com.xxl.job.core.enums.ExecutorBlockStrategyEnum
        "executorTimeout":0,                        // 任务超时时间，单位秒，大于零时生效
        "logId":1,                                  // 本次调度日志ID
        "logDateTime":1586629003729,                // 本次调度日志时间
        "glueType":"BEAN",                          // 任务模式，可选值参考 com.xxl.job.core.glue.GlueTypeEnum
        "glueSource":"xxx",                         // GLUE脚本代码
        "glueUpdatetime":1586629003727,             // GLUE脚本更新时间，用于判定脚本是否变更以及是否需要刷新
        "broadcastIndex":0,                         // 分片参数：当前分片
        "broadcastTotal":0                          // 分片参数：总分片
    }
    """
    data = await request.json()
    try:
        await request.app["executor"].run_job(RunData(**data))
    except error.JobDuplicateError as e:
        return web.json_response(dict(code=500, msg=e.message))
    except error.JobNotFoundError as e:
        return web.json_response(dict(code=500, msg=e.message))

    return web.json_response(dict(code=200, msg=None))


@routes.post("/kill")
async def kill(request: web.Request) -> web.Response:
    data = await request.json()
    await request.app["executor"].cancel_job(data["jobId"])
    return web.json_response(dict(code=200, msg=None))


# todo
@routes.post("/log")
async def log(request: web.Request) -> web.Response:
    """
        {
        "logDateTim":0,     // 本次调度日志时间
        "logId":0,          // 本次调度日志ID
        "fromLineNum":0     // 日志开始行号，滚动加载日志
    }
    """
    data = await request.json()
    logger.info("log %s" % data)
    response = {
        "code": 200,
        "msg": None,
        "content": {"fromLineNum": 1, "toLineNum": 1, "logContent": "xxx", "isEnd": True},
    }

    return web.json_response(response)


def create_app() -> web.Application:
    """
    xxl_admin_baseurl xxl调度中心的接口地址，如http://localhost:8080/xxl-job-admin/api/
    executor_names 执行器名称列表
    executor_baseurl 执行器http接口的地址 如http://172.17.0.1:9999
    """
    app = web.Application()
    app.add_routes(routes)
    return app



import asyncio
import time

import pytest

from pyxxl.enum import executorBlockStrategy
from pyxxl.error import JobDuplicateError, JobNotFoundError
from pyxxl.execute import Executor, JobHandler
from pyxxl.schema import RunData


job_handler = JobHandler()


@job_handler.register
async def pytest_executor_5():
    await asyncio.sleep(5)
    return "成功30"


@job_handler.register
async def pytest_executor_error():
    assert 1 == 2


@job_handler.register
def pytest_executor_3():
    time.sleep(3)
    return "成功30"


@pytest.mark.asyncio
async def test_runner_not_found(executor: Executor):
    executor.reset_handler(job_handler)
    with pytest.raises(JobNotFoundError):
        await executor.run_job(
            RunData(
                **dict(
                    logId=211,
                    jobId=211,
                    executorHandler="not_found",
                    executorBlockStrategy=executorBlockStrategy.DISCARD_LATER.value,
                )
            )
        )
    await executor.graceful_close()


@pytest.mark.asyncio
async def test_runner_callback(executor: Executor):
    executor.reset_handler(job_handler)
    data = RunData(
        **dict(
            logId=1,
            jobId=11,
            executorHandler="pytest_executor_5",
            executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
        )
    )
    await executor.run_job(data)
    data = RunData(
        **dict(
            logId=2,
            jobId=12,
            executorHandler="pytest_executor_error",
            executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
        )
    )
    await executor.run_job(data)
    await executor.graceful_close()
    assert executor.xxl_client.callback_result.get(1) == 200
    assert executor.xxl_client.callback_result.get(2) == 500
    assert executor.xxl_client.callback_result.get(3) is None


@pytest.mark.asyncio
async def test_runner_cancel(executor: Executor):
    executor.reset_handler(job_handler)
    await executor.run_job(
        RunData(
            **dict(
                logId=11,
                jobId=11,
                executorHandler="pytest_executor_5",
                executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
            )
        )
    )
    await executor.run_job(
        RunData(
            **dict(
                logId=12,
                jobId=12,
                executorHandler="pytest_executor_5",
                executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
            )
        )
    )
    await executor.cancel_job(11)
    await executor.graceful_close()
    assert executor.xxl_client.callback_result.get(11) is None
    assert executor.xxl_client.callback_result.get(12) == 200


@pytest.mark.asyncio
async def test_runner_SERIAL_EXECUTION(executor: Executor):
    executor.reset_handler(job_handler)
    jobId = 11
    await executor.run_job(
        RunData(
            **dict(
                logId=11,
                jobId=jobId,
                executorHandler="pytest_executor_3",
                executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
            )
        )
    )
    await executor.run_job(
        RunData(
            **dict(
                logId=21,
                jobId=jobId,
                executorHandler="pytest_executor_3",
                executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
            )
        )
    )
    await executor.run_job(
        RunData(
            **dict(
                logId=31,
                jobId=jobId,
                executorHandler="pytest_executor_3",
                executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
            )
        )
    )
    assert len(executor.queue.get(jobId)) == 2
    await executor.graceful_close()
    assert len(executor.queue.get(jobId)) == 0
    assert executor.xxl_client.callback_result.get(31) == 200


@pytest.mark.asyncio
async def test_runner_DISCARD_LATER(executor: Executor):
    executor.reset_handler(job_handler)
    jobId = 31
    executor.xxl_client.callback_result.clear()
    await executor.run_job(
        RunData(
            **dict(
                logId=11,
                jobId=jobId,
                executorHandler="pytest_executor_3",
                executorBlockStrategy=executorBlockStrategy.DISCARD_LATER.value,
            )
        )
    )
    with pytest.raises(JobDuplicateError):
        await executor.run_job(
            RunData(
                **dict(
                    logId=21,
                    jobId=jobId,
                    executorHandler="pytest_executor_3",
                    executorBlockStrategy=executorBlockStrategy.DISCARD_LATER.value,
                )
            )
        )
    await executor.graceful_close()
    assert executor.xxl_client.callback_result.get(11) == 200
    assert executor.xxl_client.callback_result.get(21) is None


@pytest.mark.asyncio
async def test_runner_COVER_EARLY(executor: Executor):
    executor.reset_handler(job_handler)
    jobId = 41
    executor.xxl_client.callback_result.clear()
    await executor.run_job(
        RunData(
            **dict(
                logId=40,
                jobId=jobId,
                executorHandler="pytest_executor_3",
                executorBlockStrategy=executorBlockStrategy.COVER_EARLY.value,
            )
        )
    )
    await executor.run_job(
        RunData(
            **dict(
                logId=41,
                jobId=jobId,
                executorHandler="pytest_executor_3",
                executorBlockStrategy=executorBlockStrategy.COVER_EARLY.value,
            )
        )
    )
    await executor.graceful_close()
    assert executor.xxl_client.callback_result.get(41) == 200
    assert executor.xxl_client.callback_result.get(40) is None



import asyncio
import functools
import logging
import time

from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Callable, Dict, List, Optional

from pyxxl import error
from pyxxl.ctx import g
from pyxxl.enum import executorBlockStrategy
from pyxxl.schema import HandlerInfo, RunData
from pyxxl.types import DecoratedCallable
from pyxxl.xxl_client import XXL


logger = logging.getLogger("pyxxl")


class JobHandler:
    _handlers: Dict[str, HandlerInfo] = {}

    def register(
        self, *args: Any, name: Optional[str] = None, replace: bool = False
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        """将函数注册到可执行的job中,如果其他地方要调用该方法,replace修改为True"""

        def func_wrapper(func: DecoratedCallable) -> DecoratedCallable:
            handler_name = name or func.__name__
            if handler_name in self._handlers and replace is False:
                raise error.JobRegisterError("handler %s already registered." % handler_name)
            self._handlers[handler_name] = HandlerInfo(handler=func)
            logger.info("register job %s,is async: %s" % (handler_name, asyncio.iscoroutinefunction(func)))

            # if asyncio.iscoroutinefunction(func):

            #     @functools.wraps(func)
            #     async def inner_wrapper(*args: Any, **kwargs: Any) -> Any:
            #         return await func(*args, **kwargs)

            # else:

            @functools.wraps(func)
            def inner_wrapper(*args: Any, **kwargs: Any) -> Any:
                return func(*args, **kwargs)

            return inner_wrapper

        if len(args) == 1:
            return func_wrapper(args[0])

        return func_wrapper

    def get(self, name: str) -> Optional[HandlerInfo]:
        return self._handlers.get(name, None)

    def handlers(self) -> List[str]:
        return list(self._handlers.keys())


class Executor:
    def __init__(
        self,
        xxl_client: XXL,
        *,
        handler: Optional[JobHandler] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        max_workers: int = 20,
        task_timeout: int = 60 * 60,
        max_queue_length: int = 30,
    ) -> None:
        self.xxl_client = xxl_client
        self.loop = loop or asyncio.get_event_loop()
        self.tasks: Dict[int, asyncio.Task] = {}
        self.queue: Dict[int, List[RunData]] = defaultdict(list)
        self.lock = asyncio.Lock()
        self.handler: JobHandler = handler or JobHandler()
        self.thread_pool = ThreadPoolExecutor(max_workers=max_workers, thread_name_prefix="pyxxl_pool")
        self.task_timeout = task_timeout
        # 串行队列的最长长度
        self.max_queue_length = max_queue_length

    async def shutdown(self) -> None:
        for _, task in self.tasks.items():
            task.cancel()

    async def run_job(self, run_data: RunData) -> None:
        handler_obj = self.handler.get(run_data.executorHandler)
        if not handler_obj:
            logger.warning("handler %s not found." % run_data.executorHandler)
            raise error.JobNotFoundError("handler %s not found." % run_data.executorHandler)

        # 一个执行器同时只能执行一个jobId相同的任务
        async with self.lock:
            current_task = self.tasks.get(run_data.jobId)
            if current_task:
                # 不用的阻塞策略
                # pylint: disable=no-else-raise
                if run_data.executorBlockStrategy == executorBlockStrategy.DISCARD_LATER.value:
                    raise error.JobDuplicateError(
                        "The same job [%s] is already executing and this has been discarded." % run_data.jobId
                    )
                elif run_data.executorBlockStrategy == executorBlockStrategy.COVER_EARLY.value:
                    logger.warning("job %s is  COVER_EARLY, logId %s replaced." % (run_data.jobId, run_data.logId))
                    await self._cancel(run_data.jobId)
                elif run_data.executorBlockStrategy == executorBlockStrategy.SERIAL_EXECUTION.value:

                    if len(self.queue[run_data.jobId]) > self.max_queue_length:
                        msg = "job %s is  SERIAL, queue length more than %s. logId %s  discard !" % (
                            run_data.jobId,
                            self.max_queue_length,
                            run_data.logId,
                        )
                        logger.error(msg)
                        raise error.JobDuplicateError(msg)
                    else:
                        queue = self.queue[run_data.jobId]
                        logger.info(
                            "job %s is in queen, logId %s wait for %s..."
                            % (run_data.jobId, run_data.logId, len(queue) + 1)
                        )
                        queue.append(run_data)
                        return
                else:
                    raise error.JobDuplicateError(
                        "unknown executorBlockStrategy [%s]." % run_data.executorBlockStrategy
                    )

            start_time = int(time.time()) * 1000
            task = self.loop.create_task(self._run(handler_obj, start_time, run_data))
            self.tasks[run_data.jobId] = task

    async def cancel_job(self, job_id: int) -> None:
        async with self.lock:
            await self._cancel(job_id)

    async def is_running(self, job_id: int) -> bool:
        return job_id in self.tasks

    async def _run(self, handler: HandlerInfo, start_time: int, data: RunData) -> None:
        try:
            g.set_xxl_run_data(data)
            logger.info("start job %s %s" % (data.jobId, data))
            func = (
                handler.handler()
                if handler.is_async
                else self.loop.run_in_executor(
                    self.thread_pool,
                    handler.handler,
                )
            )
            result = await asyncio.wait_for(func, self.task_timeout)
            logger.info("end job %s %s" % (data.jobId, data))
            await self.xxl_client.callback(data.logId, start_time, code=200, msg=result)
        except asyncio.CancelledError as e:
            logger.warning(e, exc_info=True)
            await self.xxl_client.callback(data.logId, start_time, code=500, msg="CancelledError")
        except Exception as err:  # pylint: disable=broad-except
            logger.exception(err)
            await self.xxl_client.callback(data.logId, start_time, code=500, msg=str(err))
        finally:
            await self._finish(data.jobId)

    async def _finish(self, job_id: int) -> None:
        self.tasks.pop(job_id, None)
        # 如果有队列中的任务，开始执行队列中的任务
        queue = self.queue[job_id]
        if queue:
            kwargs: RunData = queue.pop(0)
            logger.info("job %s in queue[%s], start job with logId %s" % (kwargs.jobId, len(queue), kwargs.logId))
            await self.run_job(kwargs)

    async def _cancel(self, job_id: int) -> None:
        task = self.tasks.pop(job_id, None)
        if task:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                logger.warning("Job %s cancelled." % job_id)

    async def graceful_close(self, timeout: int = 60) -> None:
        async def _graceful_close() -> None:
            while len(self.tasks) > 0:
                await asyncio.wait(self.tasks.values())

        await asyncio.wait_for(_graceful_close(), timeout=timeout)

    def reset_handler(self, handler: JobHandler) -> None:
        self.handler = handler



class XXLRegisterError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class JobDuplicateError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class JobNotFoundError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class JobRegisterError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class ClientError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)



import time

import pytest

from aiohttp.test_utils import TestClient


# import logging

# logger = logging.getLogger("pyxxl")
# handler = logging.StreamHandler()
# logger.addHandler(handler)
# logger.setLevel(logging.DEBUG)


async def _send_demoJobHandler(cli: TestClient, **kwargs):
    job_data = {
        "jobId": int(time.time() * 1000),
        "executorHandler": "demoJobHandler",
        "executorParams": "demoJobHandler",
        "executorBlockStrategy": "COVER_EARLY",
        "executorTimeout": 0,
        "logId": int(time.time() * 1000),
        "logDateTime": 1586629003729,
        "glueType": "BEAN",
        "glueSource": "xxx",
        "glueUpdatetime": 1586629003727,
        "broadcastIndex": 0,
        "broadcastTotal": 0,
    }
    job_data.update(kwargs)
    resp = await cli.post("/run", json=job_data)
    return resp, job_data["jobId"]


@pytest.mark.asyncio
async def test_run(cli: TestClient):
    resp, _ = await _send_demoJobHandler(cli, executorBlockStrategy="DISCARD_LATER", jobId=100)
    assert resp.status == 200
    assert await resp.json() == {"code": 200, "msg": None}
    # error
    resp, _ = await _send_demoJobHandler(cli, executorBlockStrategy="DISCARD_LATER", jobId=100)
    assert resp.status == 200
    response_dict = await resp.json()
    assert response_dict["code"] == 500
    assert "already executing" in response_dict["msg"]


@pytest.mark.asyncio
async def test_run_not_found(cli: TestClient):
    resp, _ = await _send_demoJobHandler(cli, executorHandler="test_run_not_found")
    assert resp.status == 200
    response_dict = await resp.json()
    assert response_dict["code"] == 500
    assert "not found" in response_dict["msg"]


@pytest.mark.asyncio
async def test_beat(cli: TestClient):
    resp = await cli.post("/beat")
    assert resp.status == 200
    assert await resp.json() == {"code": 200, "msg": None}


@pytest.mark.asyncio
async def test_idle_beat(cli: TestClient):
    resp = await cli.post("/idleBeat", json={"jobId": 1})
    assert resp.status == 200
    assert await resp.json() == {"code": 200, "msg": None}

    resp, jobId = await _send_demoJobHandler(cli, jobId=300)
    resp, jobId = await _send_demoJobHandler(cli, jobId=300)
    resp = await cli.post("/idleBeat", json={"jobId": jobId})
    response_data = await resp.json()
    assert response_data["code"] == 500
    assert response_data["msg"] == "job %s is running." % jobId


@pytest.mark.asyncio
async def test_kill(cli: TestClient):
    resp, jobId = await _send_demoJobHandler(cli)
    resp = await cli.post("/kill", json={"jobId": jobId})
    assert await resp.json() == {"code": 200, "msg": None}



from asyncio import iscoroutinefunction
from dataclasses import dataclass
from typing import Callable, Optional


@dataclass
class HandlerInfo:
    handler: Callable

    @property
    def is_async(self) -> bool:
        return iscoroutinefunction(self.handler)


@dataclass(frozen=True)
class RunData:
    jobId: int
    logId: int
    executorHandler: str
    executorBlockStrategy: str

    executorParams: Optional[str] = None
    executorTimeout: Optional[int] = None
    logDateTime: Optional[int] = None
    glueType: Optional[str] = None
    glueSource: Optional[str] = None
    glueUpdatetime: Optional[int] = None
    broadcastIndex: Optional[int] = None
    broadcastTotal: Optional[int] = None



import asyncio
import logging

from typing import Any, Dict, List, Optional, Union

import aiohttp

from pyxxl.error import ClientError, XXLRegisterError


logger = logging.getLogger("pyxxl")

JsonType = Union[None, int, str, bool, List[Any], Dict[Any, Any]]


class Response:
    def __init__(self, code: int, msg: Optional[str] = None, **kwargs: Any) -> None:
        self.code = code
        self.msg = msg

    @property
    def ok(self) -> bool:
        return self.code == 200


class XXL:
    def __init__(
        self,
        admin_url: str,
        token: Optional[str] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        retry_times: int = 0,
        retry_interval: int = 5,
        **kwargs: Any,
    ) -> None:
        self.loop = loop or asyncio.get_event_loop()
        kwargs["loop"] = self.loop
        # https://docs.aiohttp.org/en/stable/client_reference.html#baseconnector
        self.conn = aiohttp.TCPConnector(**kwargs)
        self.session: aiohttp.ClientSession = aiohttp.ClientSession(connector=self.conn)
        if not (admin_url.startswith("http") and admin_url.endswith("/")):
            raise ValueError("admin_url must like http://localhost:8080/xxl-job-admin/api/")
        self.admin_url = admin_url
        self.retry_times = retry_times
        self.retry_interval = retry_interval
        self.headers = {"XXL-JOB-ACCESS-TOKEN": token} if token else {}

    async def registry(self, key: str, value: str) -> None:
        payload = dict(registryGroup="EXECUTOR", registryKey=key, registryValue=value)
        try:
            await self._post("registry", payload, retry_times=1)
            logger.debug("registry successful. %s" % payload)
        except XXLRegisterError as e:
            logger.error("registry executor failed. %s", e.message)

    async def registryRemove(self, key: str, value: str) -> None:
        payload = dict(registryGroup="EXECUTOR", registryKey=key, registryValue=value)
        await self._post("registryRemove", payload)
        logger.info("registryRemove successful. %s" % payload)

    async def callback(self, log_id: int, timestamp: int, code: int = 200, msg: str = None) -> None:
        payload = [
            {
                "logId": log_id,
                "logDateTim": timestamp,
                "handleCode": code,
                "handleMsg": msg,
            }
        ]
        await self._post("callback", payload)
        logger.debug("callback successful. %s" % payload)

    async def _post(self, path: str, payload: JsonType, retry_times: Optional[int] = None) -> Response:
        times = 1
        retry_times = retry_times or self.retry_times
        while times <= retry_times or retry_times == 0:
            try:
                async with self.session.post(self.admin_url + path, json=payload, headers=self.headers) as response:
                    if response.status == 200:
                        r = Response(**(await response.json()))
                        if not r.ok:
                            raise XXLRegisterError(r.msg or "")
                        return r
                    raise XXLRegisterError(await response.text())
            except aiohttp.ClientConnectionError as e:
                logger.error(f"Connection error {times} times: {str(e)}, retry afert {self.retry_interval}")
                await asyncio.sleep(self.retry_interval)
                times += 1
        raise ClientError("Connection error, retry times {}".format(times))

    async def close(self) -> None:
        await self.session.close()
        logger.info("http session is closed.")



import pytest

from pyxxl.error import JobRegisterError
from pyxxl.execute import Executor


# pylint: disable=function-redefined
@pytest.mark.asyncio
async def test_hander_error(executor: Executor):
    with pytest.raises(JobRegisterError):

        @executor.handler.register
        def test_dup_error():
            ...

        @executor.handler.register
        def test_dup_error():  # noqa: F811
            ...


# pylint: disable=function-redefined
@pytest.mark.asyncio
async def test_hander(executor: Executor):
    @executor.handler.register
    def text_hander1():
        ...

    @executor.handler.register(replace=True)
    async def text_hander1():  # noqa: F811
        ...

    @executor.handler.register(name="text_hander_dup")
    def text_hander1():  # noqa: F811
        ...



import pytest

from pyxxl.ctx import g
from pyxxl.enum import executorBlockStrategy
from pyxxl.execute import Executor
from pyxxl.schema import RunData


@pytest.mark.asyncio
async def test_runner_callback(executor: Executor):
    @executor.handler.register
    async def text_ctx():
        logId = g.xxl_run_data.logId
        assert logId == 1

    data = RunData(
        **dict(
            logId=1,
            jobId=11,
            executorHandler="text_ctx",
            executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
        )
    )
    await executor.run_job(data)
    await executor.graceful_close()
    assert executor.xxl_client.callback_result.get(1) == 200



import asyncio

from typing import Generator

import pytest
import pytest_asyncio

from aiohttp.web import Application
from pytest_aiohttp.plugin import AiohttpClient, TestClient

from pyxxl.execute import Executor
from pyxxl.tests.utils import MokePyxxlRunner, MokeXXL


@pytest.fixture(scope="session")
def event_loop() -> Generator:
    """Create an instance of the default event loop for each test case."""
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
async def executor() -> Executor:
    return Executor(MokeXXL("http://localhost:8080/xxl-job-admin/api/"))


@pytest.fixture(scope="session")
def web_app() -> Application:
    runner = MokePyxxlRunner(
        "http://localhost:8080/xxl-job-admin/api/",
        executor_name="xxl-job-executor-sample",
        port=9999,
        host="127.0.0.1",
    )

    @runner.handler.register(name="demoJobHandler")
    async def test_task() -> None:
        await asyncio.sleep(60)

    return runner.create_server_app()


@pytest_asyncio.fixture
async def cli(aiohttp_client: AiohttpClient, web_app: Application) -> TestClient:
    return await aiohttp_client(web_app)



import pytest

from pyxxl.xxl_client import XXL


@pytest.mark.asyncio
async def test_param_admin_url():
    XXL("http://localhost:8080/xxl-job-admin/api/")
    XXL("https://localhost:8080/xxl-job-admin/api/")

    with pytest.raises(ValueError):
        XXL("htp://localhost:8080/xxl-job-admin/api/")

    with pytest.raises(ValueError):
        XXL("http://localhost:8080/xxl-job-admin/api")



import asyncio
import logging

from multiprocessing import Process
from typing import AsyncGenerator, Optional

from aiohttp import web

from pyxxl.execute import Executor, JobHandler
from pyxxl.service import create_app
from pyxxl.utils import ensure_host
from pyxxl.xxl_client import XXL


logger = logging.getLogger("pyxxl.run")


class PyxxlRunner:
    xxl_client: Optional[XXL] = None
    executor: Optional[Executor] = None
    register_task: Optional[asyncio.Task] = None
    daemon: Optional[Process] = None

    def __init__(
        self,
        xxl_admin_baseurl: str,
        executor_name: str,
        handler: Optional[JobHandler] = None,
        access_token: Optional[str] = None,
        host: Optional[str] = None,
        port: int = 9999,
    ):
        """

        Args:
            xxl_admin_baseurl (str): xxl-admin服务端暴露的restful接口url(如http://localhost:8080/xxl-job-admin/api/)
            executor_name (str): xxl-admin上定义的执行器名称,必须一致否则无法注册(如xxl-job-executor-sample)
            handler (JobHandler, optional): 执行器支持的job,没有预先定义的job名称也会执行失败
            access_token (str, optional): xxl-admin的认证token,如果没有开启不需要传. Defaults to None.
            host (_type_, optional): 执行器绑定的host,xxl-admin通过这个host来回调pyxxl执行器,如果不填会默认取第一个网卡的地址. Defaults to None.
            port (int, optional): 执行器绑定的http服务的端口,作用同host. Defaults to 9999.
        """
        self.host = ensure_host(host)
        self.port = port
        self.xxl_admin_baseurl = xxl_admin_baseurl
        self.executor_name = executor_name
        self.executor_baseurl = "http://{host}:{port}".format(host=self.host, port=self.port)
        self.access_token = access_token
        self.handler = handler or JobHandler()

    async def _register_task(self, xxl_client: XXL) -> None:
        # todo: 这是个调度器的bug，必须循环去注册，不然会显示为离线
        # https://github.com/xuxueli/xxl-job/issues/2090
        while True:
            await xxl_client.registry(self.executor_name, self.executor_baseurl)
            await asyncio.sleep(10)

    def _get_xxl_clint(self) -> XXL:
        """for moke"""
        return XXL(self.xxl_admin_baseurl, token=self.access_token)

    async def _init(self) -> None:
        self.xxl_client = self._get_xxl_clint()
        self.executor = Executor(self.xxl_client, handler=self.handler)
        self.register_task = asyncio.create_task(self._register_task(self.xxl_client), name="pyxxl-register")

    async def _cleanup_ctx(self, app: web.Application) -> AsyncGenerator:
        await self._init()
        app["xxl_client"] = self.xxl_client
        app["executor"] = self.executor
        app["register_task"] = self.register_task
        if self.executor and self.executor.handler:
            logger.info("register with handlers %s", list(self.executor.handler.handlers()))
        else:
            logger.warning("register with handlers is empty")

        yield

        app["register_task"].cancel()
        await app["xxl_client"].registryRemove(self.executor_name, self.executor_baseurl)
        await app["executor"].shutdown()
        await app["xxl_client"].close()
        logger.info("cleanup executor success.")

    def create_server_app(self) -> web.Application:
        app = create_app()
        app.cleanup_ctx.append(self._cleanup_ctx)
        return app

    def run_executor(self, handle_signals: bool = True) -> None:
        web.run_app(
            self.create_server_app(),
            port=self.port,
            host=self.host,
            handle_signals=handle_signals,
        )

    def run_with_daemon(self) -> None:
        def _runner() -> None:
            self.run_executor(handle_signals=True)

        daemon = Process(target=_runner, name="pyxxl", daemon=True)
        daemon.start()
        self.daemon = daemon

    # def exit_daemon(self):
    #     logger.info("Exit daemon name=%s", self.daemon.name )
    #     self.daemon.terminate()



import socket

from typing import Optional


def get_network_ip() -> str:
    """获取本机地址,会获取首个网络地址"""
    _, _, ipaddrlist = socket.gethostbyname_ex(socket.gethostname())
    return ipaddrlist[0]


def ensure_host(host: Optional[str] = None) -> str:
    return host or get_network_ip()



from enum import Enum


class executorBlockStrategy(Enum):
    SERIAL_EXECUTION = "SERIAL_EXECUTION"  # 单机串行
    DISCARD_LATER = "DISCARD_LATER"  # 丢弃后续调度
    COVER_EARLY = "COVER_EARLY"  # 关闭上次执行改为这次执行，不推荐



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyxxl', 'pyxxl.tests', 'pyxxl.tests.api']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0']

setup_kwargs = {
    'name': 'pyxxl',
    'version': '0.1.5',
    'description': 'A Python executor for XXL-jobs',
    'long_description': 'xxl-jobs 的python客户端实现\n=============================\n\n.. image:: https://img.shields.io/pypi/v/pyxxl?color=%2334D058&label=pypi%20package\n    :target: https://pypi.org/project/pyxxl\n\n.. image:: https://img.shields.io/pypi/pyversions/pyxxl.svg?color=%2334D058\n    :target: https://pypi.org/project/pyxxl\n\n.. image:: https://img.shields.io/codecov/c/github/fcfangcc/pyxxl?color=%2334D058\n    :target: https://codecov.io/gh/fcfangcc/pyxxl\n\n\n使用pyxxl可以方便的把Python写的方法注册到xxl-job中,使用xxl-job-admin管理Python定时任务和周期任务\n\n如何使用\n=======================\n.. code:: shell\n\n    pip install pyxxl\n\n具体可以查看example文件夹下面的2个例子\n\n.. code:: python\n\n    import logging\n    import asyncio\n\n    from pyxxl import PyxxlRunner, JobHandler\n\n    logger = logging.getLogger("pyxxl")\n    logger.setLevel(logging.INFO)\n    xxl_handler = JobHandler()\n\n    @xxl_handler.register\n    async def test_task():\n        await asyncio.sleep(30)\n        return "成功30"\n\n\n    @xxl_handler.register(name="xxxxx")\n    @xxxxx # 自己定义的装饰器必须放在下面\n    async def abc():\n        await asyncio.sleep(3)\n        return "成功3"\n\n\n    runner = PyxxlRunner(\n        "http://localhost:8080/xxl-job-admin/api/",\n        executor_name="xxl-job-executor-sample",\n        port=9999,\n        host="172.17.0.1",\n        handler=xxl_handler,\n    )\n    runner.run_executor()\n\n\n\n\n开发人员\n=======================\n下面是开发人员如何快捷的搭建开发调试环境\n\n=====================\n启动xxl的调度中心\n=====================\n\n.. code:: shell\n\n    ./init_dev_env.sh\n\n=====================\n启动执行器\n=====================\n.. code:: shell\n\n    poetry install\n    # 修改app.py中相关的配置信息,然后启动\n    poetry run python example/app.py\n\n\n======================\nTODOs\n======================\n\n- [x] 自定义查看日志函数\n- [x] docs\n',
    'author': 'fcfangcc',
    'author_email': 'swjfc22@live.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/fcfangcc/pyxxl',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
